# Copyright (C) 2012 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
"""
Models are data transfer objects or "light-weight" domain objects with no appreciable logic
other than their own validation. Models are used to pass data between layers of the stack. In
many ways they are similar to ORM objects.  Unlike ORM objects, they are NOT themselves allowed
to load data from a remote resource.  Thought of another way, they cannot import from
``conda.gateways``, but rather ``conda.gateways`` imports from ``conda.models`` as appropriate
to create model objects from remote resources.

Conda modules importable from ``conda.models`` are

- ``conda._vendor``
- ``conda.common``
- ``conda.models``

"""



# Copyright (C) 2012 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
from os.path import exists, join
from uuid import uuid4

import pytest

from conda.base.context import conda_tests_ctxt_mgmt_def_pol, context
from conda.common.io import dashlist, env_var, env_vars
from conda.common.serialize import yaml_round_trip_load
from conda.core.prefix_data import PrefixData
from conda.models.enums import PackageType
from conda.models.match_spec import MatchSpec
from conda.testing.integration import PYTHON_BINARY
from conda.testing.integration import Commands as CondaCommands
from conda.testing.integration import run_command as run_conda_command
from tests.test_utils import is_prefix_activated_PATHwise

from . import support_file
from .utils import Commands, make_temp_envs_dir, run_command


def package_is_installed(prefix, spec, pip=None):
    spec = MatchSpec(spec)
    prefix_recs = tuple(PrefixData(prefix, pip_interop_enabled=pip).query(spec))
    if len(prefix_recs) > 1:
        raise AssertionError(
            "Multiple packages installed.%s"
            % (dashlist(prec.dist_str() for prec in prefix_recs))
        )
    is_installed = bool(len(prefix_recs))
    if is_installed and pip is True:
        assert prefix_recs[0].package_type in (
            PackageType.VIRTUAL_PYTHON_WHEEL,
            PackageType.VIRTUAL_PYTHON_EGG_MANAGEABLE,
            PackageType.VIRTUAL_PYTHON_EGG_UNMANAGEABLE,
            PackageType.VIRTUAL_PYTHON_EGG_LINK,
        )
    if is_installed and pip is False:
        assert prefix_recs[0].package_type in (
            None,
            PackageType.NOARCH_GENERIC,
            PackageType.NOARCH_PYTHON,
        )
    return is_installed


def get_env_vars(prefix):
    pd = PrefixData(prefix)

    env_vars = pd.get_environment_env_vars()

    return env_vars


@pytest.mark.integration
def test_create_update():
    with make_temp_envs_dir() as envs_dir:
        with env_var(
            "CONDA_ENVS_DIRS",
            envs_dir,
            stack_callback=conda_tests_ctxt_mgmt_def_pol,
        ):
            env_name = str(uuid4())[:8]
            prefix = join(envs_dir, env_name)
            python_path = join(prefix, PYTHON_BINARY)

            run_command(
                Commands.CREATE,
                env_name,
                support_file("example/environment_pinned.yml"),
            )
            assert exists(python_path)
            assert package_is_installed(prefix, "flask=2.0.2")

            env_vars = get_env_vars(prefix)
            assert env_vars["FIXED"] == "fixed"
            assert env_vars["CHANGES"] == "original_value"
            assert env_vars["GETS_DELETED"] == "not_actually_removed_though"
            assert env_vars.get("NEW_VAR") is None

            run_command(
                Commands.UPDATE,
                env_name,
                support_file("example/environment_pinned_updated.yml"),
            )
            assert package_is_installed(prefix, "flask=2.0.3")
            assert not package_is_installed(prefix, "flask=2.0.2")

            env_vars = get_env_vars(prefix)
            assert env_vars["FIXED"] == "fixed"
            assert env_vars["CHANGES"] == "updated_value"
            assert env_vars["NEW_VAR"] == "new_var"

            # This ends up sticking around since there is no real way of knowing that an environment
            # variable _used_ to be in the variables dict, but isn't any more.
            assert env_vars["GETS_DELETED"] == "not_actually_removed_though"


@pytest.mark.skip(reason="Need to find an appropriate server to test this on.")
@pytest.mark.integration
def test_create_host_port():
    with make_temp_envs_dir() as envs_dir:
        with env_var(
            "CONDA_ENVS_DIRS",
            envs_dir,
            stack_callback=conda_tests_ctxt_mgmt_def_pol,
        ):
            env_name = str(uuid4())[:8]
            prefix = join(envs_dir, env_name)
            python_path = join(prefix, PYTHON_BINARY)

            run_command(
                Commands.CREATE,
                env_name,
                support_file("example/environment_host_port.yml"),
            )
            assert exists(python_path)
            assert package_is_installed(prefix, "flask=2.0.3")


# This test will not run from an unactivated conda in an IDE. You *will* get complaints about being unable
# to load the SSL module. Never try to test conda from outside an activated env. Maybe this should be a
# session fixture with autouse=True so we just refuse to run the testsuite in that case?!
@pytest.mark.skipif(
    not is_prefix_activated_PATHwise(),
    reason="You are running `pytest` outside of proper activation. "
    "The entries necessary for conda to operate correctly "
    "are not on PATH.  Please use `conda activate`",
)
@pytest.mark.integration
def test_create_advanced_pip():
    with make_temp_envs_dir() as envs_dir:
        with env_vars(
            {
                "CONDA_ENVS_DIRS": envs_dir,
            },
            stack_callback=conda_tests_ctxt_mgmt_def_pol,
        ):
            env_name = str(uuid4())[:8]
            prefix = join(envs_dir, env_name)
            python_path = join(prefix, PYTHON_BINARY)

            run_command(
                Commands.CREATE,
                env_name,
                support_file(join("advanced-pip", "environment.yml")),
            )
            assert exists(python_path)
            PrefixData._cache_.clear()
            assert package_is_installed(prefix, "argh", pip=True)
            assert package_is_installed(
                prefix, "module-to-install-in-editable-mode", pip=True
            )
            try:
                assert package_is_installed(prefix, "six", pip=True)
            except AssertionError:
                # six may now be conda-installed because of packaging changes
                assert package_is_installed(prefix, "six", pip=False)
            assert package_is_installed(prefix, "xmltodict=0.10.2", pip=True)


@pytest.mark.integration
def test_create_empty_env():
    with make_temp_envs_dir() as envs_dir:
        with env_var(
            "CONDA_ENVS_DIRS",
            envs_dir,
            stack_callback=conda_tests_ctxt_mgmt_def_pol,
        ):
            env_name = str(uuid4())[:8]
            prefix = join(envs_dir, env_name)
            run_command(Commands.CREATE, env_name, support_file("empty_env.yml"))
            assert exists(prefix)


@pytest.mark.skipif(
    context.subdir
    not in {"linux-64", "linux-ppc64le", "osx-64", "win-32", "win-64", "linux-32"},
    reason="Newer platforms lack Python 2",
)
@pytest.mark.integration
def test_create_env_default_packages():
    with make_temp_envs_dir() as envs_dir:
        with env_var(
            "CONDA_ENVS_DIRS",
            envs_dir,
            stack_callback=conda_tests_ctxt_mgmt_def_pol,
        ):
            # set packages
            run_conda_command(
                CondaCommands.CONFIG,
                envs_dir,
                "--add",
                "create_default_packages",
                "pip",
            )
            run_conda_command(
                CondaCommands.CONFIG,
                envs_dir,
                "--add",
                "create_default_packages",
                "flask",
            )
            stdout, stderr, _ = run_conda_command(
                CondaCommands.CONFIG, envs_dir, "--show"
            )
            yml_obj = yaml_round_trip_load(stdout)
            assert yml_obj["create_default_packages"] == ["flask", "pip"]

            assert not package_is_installed(envs_dir, "python=2")
            assert not package_is_installed(envs_dir, "pytz")
            assert not package_is_installed(envs_dir, "flask")

            env_name = str(uuid4())[:8]
            prefix = join(envs_dir, env_name)
            run_command(
                Commands.CREATE, env_name, support_file("env_with_dependencies.yml")
            )
            assert exists(prefix)
            assert package_is_installed(prefix, "python=2")
            assert package_is_installed(prefix, "pytz")
            assert package_is_installed(prefix, "flask")


@pytest.mark.skipif(
    context.subdir
    not in {"linux-64", "linux-ppc64le", "osx-64", "win-32", "win-64", "linux-32"},
    reason="Newer platforms lack Python 2",
)
@pytest.mark.integration
def test_create_env_no_default_packages():
    with make_temp_envs_dir() as envs_dir:
        with env_var(
            "CONDA_ENVS_DIRS",
            envs_dir,
            stack_callback=conda_tests_ctxt_mgmt_def_pol,
        ):
            # set packages
            run_conda_command(
                CondaCommands.CONFIG,
                envs_dir,
                "--add",
                "create_default_packages",
                "pip",
            )
            run_conda_command(
                CondaCommands.CONFIG,
                envs_dir,
                "--add",
                "create_default_packages",
                "flask",
            )
            stdout, stderr, _ = run_conda_command(
                CondaCommands.CONFIG, envs_dir, "--show"
            )
            yml_obj = yaml_round_trip_load(stdout)
            assert yml_obj["create_default_packages"] == ["flask", "pip"]

            assert not package_is_installed(envs_dir, "python=2")
            assert not package_is_installed(envs_dir, "pytz")
            assert not package_is_installed(envs_dir, "flask")

            env_name = str(uuid4())[:8]
            prefix = join(envs_dir, env_name)
            run_command(
                Commands.CREATE,
                env_name,
                support_file("env_with_dependencies.yml"),
                "--no-default-packages",
            )
            assert exists(prefix)
            assert package_is_installed(prefix, "python=2")
            assert package_is_installed(prefix, "pytz")
            assert not package_is_installed(prefix, "flask")


# removed from class to be able to accept pytest fixture
def test_create_update_remote_env_file(support_file_server_port):
    with make_temp_envs_dir() as envs_dir:
        with env_var(
            "CONDA_ENVS_DIRS", envs_dir, stack_callback=conda_tests_ctxt_mgmt_def_pol
        ):
            env_name = str(uuid4())[:8]
            prefix = join(envs_dir, env_name)
            python_path = join(prefix, PYTHON_BINARY)

            run_command(
                Commands.CREATE,
                env_name,
                support_file(
                    "example/environment_pinned.yml",
                    remote=True,
                    port=support_file_server_port,
                ),
            )
            assert exists(python_path)
            assert package_is_installed(prefix, "flask=2.0.2")

            env_vars = get_env_vars(prefix)
            assert env_vars["FIXED"] == "fixed"
            assert env_vars["CHANGES"] == "original_value"
            assert env_vars["GETS_DELETED"] == "not_actually_removed_though"
            assert env_vars.get("NEW_VAR") is None

            run_command(
                Commands.UPDATE,
                env_name,
                support_file(
                    "example/environment_pinned_updated.yml",
                    remote=True,
                    port=support_file_server_port,
                ),
            )
            assert package_is_installed(prefix, "flask=2.0.3")
            assert not package_is_installed(prefix, "flask=2.0.2")

            env_vars = get_env_vars(prefix)
            assert env_vars["FIXED"] == "fixed"
            assert env_vars["CHANGES"] == "updated_value"
            assert env_vars["NEW_VAR"] == "new_var"

            # This ends up sticking around since there is no real way of knowing that an environment
            # variable _used_ to be in the variables dict, but isn't any more.
            assert env_vars["GETS_DELETED"] == "not_actually_removed_though"



# Copyright (C) 2012 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
"""CLI implementation for `conda-env config`.

Allows for programmatically interacting with conda-env's configuration files (e.g., `~/.condarc`).
"""
from argparse import RawDescriptionHelpFormatter

from .main_vars import configure_parser as configure_vars_parser

config_description = """
Configure a conda environment
"""

config_example = """
examples:
    conda env config vars list
    conda env config --append channels conda-forge
"""


def configure_parser(sub_parsers):
    config_parser = sub_parsers.add_parser(
        "config",
        formatter_class=RawDescriptionHelpFormatter,
        description=config_description,
        help=config_description,
        epilog=config_example,
    )
    config_parser.set_defaults(func=".main_config.execute")
    config_subparser = config_parser.add_subparsers()
    configure_vars_parser(config_subparser)


def execute(args, parser):
    parser.parse_args(["config", "--help"])



# Copyright (C) 2012 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
"""Conda package installation logic.

Core logic for `conda [create|install|update|remove]` commands.

See conda.cli.main_create, conda.cli.main_install, conda.cli.main_update, and
conda.cli.main_remove for the entry points into this module.
"""
import os
from logging import getLogger
from os.path import abspath, basename, exists, isdir, isfile, join

from .. import CondaError
from ..auxlib.ish import dals
from ..base.constants import REPODATA_FN, ROOT_ENV_NAME, DepsModifier, UpdateModifier
from ..base.context import context, locate_prefix_by_name
from ..common.constants import NULL
from ..common.path import is_package_file, paths_equal
from ..core.index import calculate_channel_urls, get_index
from ..core.prefix_data import PrefixData
from ..exceptions import (
    CondaExitZero,
    CondaImportError,
    CondaOSError,
    CondaSystemExit,
    CondaValueError,
    DirectoryNotACondaEnvironmentError,
    DirectoryNotFoundError,
    DryRunExit,
    EnvironmentLocationNotFound,
    NoBaseEnvironmentError,
    PackageNotInstalledError,
    PackagesNotFoundError,
    SpecsConfigurationConflictError,
    TooManyArgumentsError,
    UnsatisfiableError,
)
from ..gateways.disk.create import mkdir_p
from ..gateways.disk.delete import delete_trash, path_is_clean
from ..misc import clone_env, explicit, touch_nonadmin
from ..models.match_spec import MatchSpec
from ..plan import revert_actions
from ..resolve import ResolvePackageNotFound
from . import common
from .common import check_non_admin

log = getLogger(__name__)
stderrlog = getLogger("conda.stderr")


def check_prefix(prefix, json=False):
    name = basename(prefix)
    error = None
    if name == ROOT_ENV_NAME:
        error = "'%s' is a reserved environment name" % name
    if exists(prefix):
        if isdir(prefix) and "conda-meta" not in tuple(
            entry.name for entry in os.scandir(prefix)
        ):
            return None
        error = "prefix already exists: %s" % prefix

    if error:
        raise CondaValueError(error, json)

    if " " in prefix:
        stderrlog.warning(
            "WARNING: A space was detected in your requested environment path:\n"
            f"'{prefix}'\n"
            "Spaces in paths can sometimes be problematic. To minimize issues,\n"
            "make sure you activate your environment before running any executables!\n"
        )


def clone(src_arg, dst_prefix, json=False, quiet=False, index_args=None):
    if os.sep in src_arg:
        src_prefix = abspath(src_arg)
        if not isdir(src_prefix):
            raise DirectoryNotFoundError(src_arg)
    else:
        src_prefix = locate_prefix_by_name(src_arg)

    if not json:
        print("Source:      %s" % src_prefix)
        print("Destination: %s" % dst_prefix)

    actions, untracked_files = clone_env(
        src_prefix, dst_prefix, verbose=not json, quiet=quiet, index_args=index_args
    )

    if json:
        common.stdout_json_success(
            actions=actions,
            untracked_files=list(untracked_files),
            src_prefix=src_prefix,
            dst_prefix=dst_prefix,
        )


def print_activate(env_name_or_prefix):  # pragma: no cover
    if not context.quiet and not context.json:
        if " " in env_name_or_prefix:
            env_name_or_prefix = f'"{env_name_or_prefix}"'
        message = dals(
            f"""
        #
        # To activate this environment, use
        #
        #     $ conda activate {env_name_or_prefix}
        #
        # To deactivate an active environment, use
        #
        #     $ conda deactivate
        """
        )
        print(message)  # TODO: use logger


def get_revision(arg, json=False):
    try:
        return int(arg)
    except ValueError:
        raise CondaValueError("expected revision number, not: '%s'" % arg, json)


def install(args, parser, command="install"):
    """Logic for `conda install`, `conda update`, and `conda create`."""
    context.validate_configuration()
    check_non_admin()
    # this is sort of a hack.  current_repodata.json may not have any .tar.bz2 files,
    #    because it deduplicates records that exist as both formats.  Forcing this to
    #    repodata.json ensures that .tar.bz2 files are available
    if context.use_only_tar_bz2:
        args.repodata_fns = ("repodata.json",)

    newenv = bool(command == "create")
    isupdate = bool(command == "update")
    isinstall = bool(command == "install")
    isremove = bool(command == "remove")
    prefix = context.target_prefix
    if newenv:
        check_prefix(prefix, json=context.json)
    if context.force_32bit and prefix == context.root_prefix:
        raise CondaValueError("cannot use CONDA_FORCE_32BIT=1 in base env")
    if isupdate and not (
        args.file
        or args.packages
        or context.update_modifier == UpdateModifier.UPDATE_ALL
    ):
        raise CondaValueError(
            """no package names supplied
# Example: conda update -n myenv scipy
"""
        )

    if not newenv:
        if isdir(prefix):
            delete_trash(prefix)
            if not isfile(join(prefix, "conda-meta", "history")):
                if paths_equal(prefix, context.conda_prefix):
                    raise NoBaseEnvironmentError()
                else:
                    if not path_is_clean(prefix):
                        raise DirectoryNotACondaEnvironmentError(prefix)
            else:
                # fall-through expected under normal operation
                pass
        else:
            if hasattr(args, "mkdir") and args.mkdir:
                try:
                    mkdir_p(prefix)
                except OSError as e:
                    raise CondaOSError(
                        "Could not create directory: %s" % prefix, caused_by=e
                    )
            else:
                raise EnvironmentLocationNotFound(prefix)

    args_packages = [s.strip("\"'") for s in args.packages]
    if newenv and not args.no_default_packages:
        # Override defaults if they are specified at the command line
        # TODO: rework in 4.4 branch using MatchSpec
        args_packages_names = [
            pkg.replace(" ", "=").split("=", 1)[0] for pkg in args_packages
        ]
        for default_pkg in context.create_default_packages:
            default_pkg_name = default_pkg.replace(" ", "=").split("=", 1)[0]
            if default_pkg_name not in args_packages_names:
                args_packages.append(default_pkg)

    index_args = {
        "use_cache": args.use_index_cache,
        "channel_urls": context.channels,
        "unknown": args.unknown,
        "prepend": not args.override_channels,
        "use_local": args.use_local,
    }

    num_cp = sum(is_package_file(s) for s in args_packages)
    if num_cp:
        if num_cp == len(args_packages):
            explicit(args_packages, prefix, verbose=not context.quiet)
            return
        else:
            raise CondaValueError(
                "cannot mix specifications with conda package" " filenames"
            )

    specs = []
    if args.file:
        for fpath in args.file:
            try:
                specs.extend(common.specs_from_url(fpath, json=context.json))
            except UnicodeError:
                raise CondaError(
                    "Error reading file, file should be a text file containing"
                    " packages \nconda create --help for details"
                )
        if "@EXPLICIT" in specs:
            explicit(specs, prefix, verbose=not context.quiet, index_args=index_args)
            return
    specs.extend(common.specs_from_args(args_packages, json=context.json))

    if isinstall and args.revision:
        get_revision(args.revision, json=context.json)
    elif isinstall and not (args.file or args_packages):
        raise CondaValueError(
            "too few arguments, " "must supply command line package specs or --file"
        )

    # for 'conda update', make sure the requested specs actually exist in the prefix
    # and that they are name-only specs
    if isupdate and context.update_modifier != UpdateModifier.UPDATE_ALL:
        prefix_data = PrefixData(prefix)
        for spec in specs:
            spec = MatchSpec(spec)
            if not spec.is_name_only_spec:
                raise CondaError(
                    "Invalid spec for 'conda update': %s\n"
                    "Use 'conda install' instead." % spec
                )
            if not prefix_data.get(spec.name, None):
                raise PackageNotInstalledError(prefix, spec.name)

    if newenv and args.clone:
        if args.packages:
            raise TooManyArgumentsError(
                0,
                len(args.packages),
                list(args.packages),
                "did not expect any arguments for --clone",
            )

        clone(
            args.clone,
            prefix,
            json=context.json,
            quiet=context.quiet,
            index_args=index_args,
        )
        touch_nonadmin(prefix)
        print_activate(args.name or prefix)
        return

    repodata_fns = args.repodata_fns
    if not repodata_fns:
        repodata_fns = context.repodata_fns
    if REPODATA_FN not in repodata_fns:
        repodata_fns.append(REPODATA_FN)

    args_set_update_modifier = (
        hasattr(args, "update_modifier") and args.update_modifier != NULL
    )
    # This helps us differentiate between an update, the --freeze-installed option, and the retry
    # behavior in our initial fast frozen solve
    _should_retry_unfrozen = (
        not args_set_update_modifier
        or args.update_modifier
        not in (UpdateModifier.FREEZE_INSTALLED, UpdateModifier.UPDATE_SPECS)
    ) and not newenv

    for repodata_fn in repodata_fns:
        try:
            if isinstall and args.revision:
                index = get_index(
                    channel_urls=index_args["channel_urls"],
                    prepend=index_args["prepend"],
                    platform=None,
                    use_local=index_args["use_local"],
                    use_cache=index_args["use_cache"],
                    unknown=index_args["unknown"],
                    prefix=prefix,
                    repodata_fn=repodata_fn,
                )
                unlink_link_transaction = revert_actions(
                    prefix, get_revision(args.revision), index
                )
            else:
                solver_backend = context.plugin_manager.get_cached_solver_backend()
                solver = solver_backend(
                    prefix,
                    context.channels,
                    context.subdirs,
                    specs_to_add=specs,
                    repodata_fn=repodata_fn,
                    command=args.cmd,
                )
                update_modifier = context.update_modifier
                if (isinstall or isremove) and args.update_modifier == NULL:
                    update_modifier = UpdateModifier.FREEZE_INSTALLED
                deps_modifier = context.deps_modifier
                if isupdate:
                    deps_modifier = context.deps_modifier or DepsModifier.UPDATE_SPECS

                unlink_link_transaction = solver.solve_for_transaction(
                    deps_modifier=deps_modifier,
                    update_modifier=update_modifier,
                    force_reinstall=context.force_reinstall or context.force,
                    should_retry_solve=(
                        _should_retry_unfrozen or repodata_fn != repodata_fns[-1]
                    ),
                )
            # we only need one of these to work.  If we haven't raised an exception,
            #   we're good.
            break

        except (ResolvePackageNotFound, PackagesNotFoundError) as e:
            if not getattr(e, "allow_retry", True):
                raise e  # see note in next except block
            # end of the line.  Raise the exception
            if repodata_fn == repodata_fns[-1]:
                # PackagesNotFoundError is the only exception type we want to raise.
                #    Over time, we should try to get rid of ResolvePackageNotFound
                if isinstance(e, PackagesNotFoundError):
                    raise e
                else:
                    channels_urls = tuple(
                        calculate_channel_urls(
                            channel_urls=index_args["channel_urls"],
                            prepend=index_args["prepend"],
                            platform=None,
                            use_local=index_args["use_local"],
                        )
                    )
                    # convert the ResolvePackageNotFound into PackagesNotFoundError
                    raise PackagesNotFoundError(e._formatted_chains, channels_urls)

        except (UnsatisfiableError, SystemExit, SpecsConfigurationConflictError) as e:
            if not getattr(e, "allow_retry", True):
                # TODO: This is a temporary workaround to allow downstream libraries
                # to inject this attribute set to False and skip the retry logic
                # Other solvers might implement their own internal retry logic without
                # depending --freeze-install implicitly like conda classic does. Example
                # retry loop in conda-libmamba-solver:
                # https://github.com/conda-incubator/conda-libmamba-solver/blob/da5b1ba/conda_libmamba_solver/solver.py#L254-L299
                # If we end up raising UnsatisfiableError, we annotate it with `allow_retry`
                # so we don't have go through all the repodatas and freeze-installed logic
                # unnecessarily (see https://github.com/conda/conda/issues/11294). see also:
                # https://github.com/conda-incubator/conda-libmamba-solver/blob/7c698209/conda_libmamba_solver/solver.py#L617
                raise e
            # Quick solve with frozen env or trimmed repodata failed.  Try again without that.
            if not hasattr(args, "update_modifier"):
                if repodata_fn == repodata_fns[-1]:
                    raise e
            elif _should_retry_unfrozen:
                try:
                    unlink_link_transaction = solver.solve_for_transaction(
                        deps_modifier=deps_modifier,
                        update_modifier=UpdateModifier.UPDATE_SPECS,
                        force_reinstall=context.force_reinstall or context.force,
                        should_retry_solve=(repodata_fn != repodata_fns[-1]),
                    )
                except (
                    UnsatisfiableError,
                    SystemExit,
                    SpecsConfigurationConflictError,
                ) as e:
                    # Unsatisfiable package specifications/no such revision/import error
                    if e.args and "could not import" in e.args[0]:
                        raise CondaImportError(str(e))
                    # we want to fall through without raising if we're not at the end of the list
                    #    of fns.  That way, we fall to the next fn.
                    if repodata_fn == repodata_fns[-1]:
                        raise e
            elif repodata_fn != repodata_fns[-1]:
                continue  # if we hit this, we should retry with next repodata source
            else:
                # end of the line.  Raise the exception
                # Unsatisfiable package specifications/no such revision/import error
                if e.args and "could not import" in e.args[0]:
                    raise CondaImportError(str(e))
                raise e
    handle_txn(unlink_link_transaction, prefix, args, newenv)


def handle_txn(unlink_link_transaction, prefix, args, newenv, remove_op=False):
    if unlink_link_transaction.nothing_to_do:
        if remove_op:
            # No packages found to remove from environment
            raise PackagesNotFoundError(args.package_names)
        elif not newenv:
            if context.json:
                common.stdout_json_success(
                    message="All requested packages already installed."
                )
            else:
                print("\n# All requested packages already installed.\n")
            return

    if not context.json:
        unlink_link_transaction.print_transaction_summary()
        common.confirm_yn()

    elif context.dry_run:
        actions = unlink_link_transaction._make_legacy_action_groups()[0]
        common.stdout_json_success(prefix=prefix, actions=actions, dry_run=True)
        raise DryRunExit()

    try:
        unlink_link_transaction.download_and_extract()
        if context.download_only:
            raise CondaExitZero(
                "Package caches prepared. UnlinkLinkTransaction cancelled with "
                "--download-only option."
            )
        unlink_link_transaction.execute()

    except SystemExit as e:
        raise CondaSystemExit("Exiting", e)

    if newenv:
        touch_nonadmin(prefix)
        print_activate(args.name or prefix)

    if context.json:
        actions = unlink_link_transaction._make_legacy_action_groups()[0]
        common.stdout_json_success(prefix=prefix, actions=actions)



# Copyright (C) 2012 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
"""Implements object describing a symbolic link from the base environment to a private environment.

Since private environments are an unrealized feature of conda and has been deprecated this data
model no longer serves a purpose and has also been deprecated.
"""
from logging import getLogger

from ..auxlib.entity import Entity, EnumField, StringField
from ..deprecations import deprecated
from .enums import LeasedPathType

log = getLogger(__name__)


@deprecated("24.3", "24.9")
class LeasedPathEntry(Entity):
    """
    _path: short path for the leased path, using forward slashes
    target_path: the full path to the executable in the private env
    target_prefix: the full path to the private environment
    leased_path: the full path for the lease in the root prefix
    package_name: the package holding the lease
    leased_path_type: application_entry_point

    """

    _path = StringField()
    target_path = StringField()
    target_prefix = StringField()
    leased_path = StringField()
    package_name = StringField()
    leased_path_type = EnumField(LeasedPathType)



# Copyright (C) 2012 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
from conda_env.env import Environment
from conda_env.specs.requirements import RequirementsSpec

from .. import support_file


def test_no_environment_file():
    spec = RequirementsSpec(name=None, filename="not-a-file")
    assert not spec.can_handle()


def test_no_name():
    spec = RequirementsSpec(filename=support_file("requirements.txt"))
    assert not spec.can_handle()


def test_req_file_and_name():
    spec = RequirementsSpec(filename=support_file("requirements.txt"), name="env")
    assert spec.can_handle()


def test_environment():
    spec = RequirementsSpec(filename=support_file("requirements.txt"), name="env")
    assert isinstance(spec.environment, Environment)
    assert spec.environment.dependencies["conda"][0] == "flask==1.1.1"
